pragma Style_Checks ("N3aAbCdefhiIklnOprStux");
pragma SPARK_Mode;
with {prefix}RFLX_Generic_Types.Generic_Operators;

package {prefix}RFLX_Types.Operators is new {prefix}RFLX_Types.Generic_Operators;
