--
--  Copyright (C) 2014-2022, AdaCore
--  SPDX-License-Identifier: Apache-2.0
--

package Langkit_Support with Pure is

   External_Name_Prefix : constant String := "langkit_support__int__";
   --  Prefix for explicit external names in Langkit_Support

end Langkit_Support;
